%% demo01_gdx_mamcost1fnntrain_main.m
% Case Study: Mamdani Neuro-Fuzzy System
%% CC BY-SA (Attribution-ShareAlike) Mamdani based neuro-fuzzy system with center-of-sets defuzzification 
%by Dr. Juan R. Castro of Universidad Autnoma de Baja California 
%is distributed under a Creative Commons Attribution-ShareAlike 4.0 International License 
%(https://creativecommons.org/licenses/by-sa/4.0/). 
%This license allows others to remix, retouch, and create from their work, even for commercial purposes, 
%as long as they give credit and license their new creations under the same terms. All new works based on 
%yours will carry the same license, so any derivative works will also allow commercial use.


clearvars;clear;clc;

%% Step 1. Loading dataset
 load simplefit_dataset;
 %%training set, n (attributes) x q (instances or data)
 inputs = simplefitInputs;
 %%target set, m (outputs) x q (instances or data)
 targets = simplefitTargets;

%% Other datasets
% load gauss3;
% inputs = xpeak';
% targets = ypeak';

% load engine_dataset
% inputs = engineInputs;
% targets = engineTargets;

% load chemical_dataset
% [inp,targets]=chemical_dataset;
% inputs=inp([1 5 7],:);

% load abalone_dataset
% [inp,targets]=abalone_dataset;
% inputs=inp([6 8],:);

% load bodyfat_dataset
% [inp,targets]=bodyfat_dataset;
% inputs=inp([2 5 6],:);

%% Step 2. Set up the percentage of training, validation and test samples that going to use in training stage.

divideParam.trainRatio = 0.60;
divideParam.valRatio    = 0.20;
divideParam.testRatio   = 0.20;
divideFcn = 'dividerand';
trainV.name = 'Training';
valV.name = 'Validation';
testV.name = 'Test';

Q=size(inputs,2);

%% Step 3. Building a FIS, the parameters required in function genfismamt1fnn are inputs, targets and rule numbers
fisB_Init = genfis1mamt1fnn(inputs', targets',5); %%Batch learning method
fisO_Init = genfis1mamt1fnn(inputs', targets',9);%%Online learning method
fisMB_Init = genfis1mamt1fnn(inputs', targets',7);%%Mini-Batch learning method

%% Building structures to use in ParFor (parallelization is used to accelerate the execution).
%% Batch
fis_Batch=cell(1,30);
outBatch_result=cell(1,30);
trainBatch=cell(1,30);

%% Online
fis_Online=cell(1,30);
outOnline_result=cell(1,30);
trainOnline=cell(1,30);

%% Mini-Batch
fis_MiniBatch=cell(1,30);
outMiniBatch_result=cell(1,30);
trainMiniBatch=cell(1,30);

tic
for i=1:30
    
    %% For each experimentation a different data sample is created randomly for training, validation and testing
    [trainV.indices,valV.indices,testV.indices] = feval(divideFcn,1:Q,divideParam);
    [trainV.X,valV.X,testV.X] = divideind(inputs,trainV.indices,valV.indices,testV.indices);
    [trainV.T,valV.T,testV.T] = divideind(targets,trainV.indices,valV.indices,testV.indices);

    trainV.Q = length(trainV.indices);
    valV.Q = length(valV.indices);
    testV.Q = length(testV.indices);

    %% Parallel execution starts
    parfor j=1:1
      trainParam=struct();
      
      %% Step 4. Hyperparameters are configured
      trainParam.epochs = 10000;
      trainParam.goal = 1e-3;
      trainParam.max_perf_inc = 1.05;
      trainParam.mc = 0.9;
      trainParam.min_grad = 1.0e-6;
      trainParam.show = 1;
      trainParam.time = inf;
      trainParam.max_fail = 20;
      trainParam.experiment=i;
      
      %% Step 5. For each specific learning method, it must be established lr (learning rate), 
      %% lr_dec ( learning rate decrement) , lr_inc ( learning rate increment).
      %% Batch
      trainParam.lr = 0.001;
      trainParam.lr_dec = 0.5;
      trainParam.lr_inc = 1.025;
      [fisBatch,trBatch] = feval('mamcost1fnntraingdx_Batch',fisB_Init,trainV,valV,testV,trainParam);
      fis_B{j}=fisBatch;
      outB = evalfis(inputs,fisBatch);

      outBatch{j}=outB;
      tr_Batch{j}=trBatch;
      
      %% Online - only for this learning method, it is necessary to establish the increment parameter
      trainParam.lr = 0.1;
      trainParam.lr_dec = 0.5;
      trainParam.lr_inc = 1.025;
      trainParam.increment= 1.005;

      [fisOnline,trOnline] = feval('mamcost1fnntraingdx_Online',fisO_Init,trainV,valV,testV,trainParam);
      fis_O{j}=fisOnline;
      outO = evalfis(inputs,fisOnline);
      outOnline{j}=outO;
      tr_Online{j}=trOnline;
        
      %% Mini-Batch - only for this learning method, it is necessary to establish the batchsize parameter
      trainParam.lr = 0.01;
      trainParam.lr_dec = 0.5;
      trainParam.lr_inc = 1.025;
      trainParam.batchSize=15;
      
      [fisMiniBatch,trMiniBatch] = feval('mamcost1fnntraingdx_MiniBatch',fisMB_Init,trainV,valV,testV,trainParam);
      fis_MB{j}=fisMiniBatch;
      outMB = evalfis(inputs,fisMiniBatch);

      outMiniBatch{j}=outMB;
      tr_MiniBatch{j}=trMiniBatch;

    end
    
    %% Each adjusted FIS and its evaluation must be stored, with the objective to find the better result of all experiment,
    %% this should done for each learning method.
    %% Batch
    fis_Batch{i}=fis_B;
    outBatch_result{i}=outBatch;
    trainBatch{i}=tr_Batch;
    
    %% Online
    fis_Online{i}=fis_O;
    outOnline_result{i}=outOnline;
    trainOnline{i}=tr_Online;
    
    %% Mini-Batch
    fis_MiniBatch{i}=fis_MB;
    outMiniBatch_result{i}=outMiniBatch;
    trainMiniBatch{i}=tr_MiniBatch;
end  
toc

out_targets=size(targets,1);

%% For each experiment the metrics are calculated. 
%% The metrics are R (coefficient of correlation) R2 (coefficient of determination), SSE (Sum of Squared Error),
%% MSE (Mean-Squared Error) and RMSE (Root Mean-Squared Error)
%% this should done for each learning method
for i=1:out_targets
  for j=1:30    
   %% Batch
   r_Batch(i,j)=regression(targets(i,:),outBatch_result{1,j}{1,1}(:,i)');
   r2_Batch{i,j} = fitlm(targets(i,:),outBatch_result{1,j}{1,1}(:,i)');
   r2_Batch_adjusted(i,j)=r2_Batch{i,j}.Rsquared.Adjusted;
   r2_Batch_SSE(i,j)=r2_Batch{i,j}.SSE;
   r2_Batch_MSE(i,j)=r2_Batch{i,j}.MSE;
   r2_Batch_RMSE(i,j)=r2_Batch{i,j}.RMSE;    
   
   %% Online 
   r_Online(i,j)=regression(targets(i,:),outOnline_result{1,j}{1,1}(:,i)');
   r2_Online{i,j} = fitlm(targets(i,:),outOnline_result{1,j}{1,1}(:,i)');
   r2_Online_adjusted(i,j)=r2_Online{i,j}.Rsquared.Adjusted;
   r2_Online_SSE(i,j)=r2_Online{i,j}.SSE;
   r2_Online_MSE(i,j)=r2_Online{i,j}.MSE;
   r2_Online_RMSE(i,j)=r2_Online{i,j}.RMSE;   
  
  %% Mini-Batch
   r_MiniBatch(i,j)=regression(targets(i,:),outMiniBatch_result{1,j}{1,1}(:,i)');
   r2_MiniBatch{i,j} = fitlm(targets(i,:),outMiniBatch_result{1,j}{1,1}(:,i)');
   r2_MiniBatch_adjusted(i,j)=r2_MiniBatch{i,j}.Rsquared.Adjusted;
   r2_MiniBatch_SSE(i,j)=r2_MiniBatch{i,j}.SSE;
   r2_MiniBatch_MSE(i,j)=r2_MiniBatch{i,j}.MSE;
   r2_MiniBatch_RMSE(i,j)=r2_MiniBatch{i,j}.RMSE;
  end
end
     
%% Building the results matrix
%% this should done for each learning method

for i=1:out_targets
  %% Batch
  R_BR(i,1:4)=[nanmin(r_Batch(i,:)) nanmean(r_Batch(i,:)) nanmax(r_Batch(i,:)) nanstd(r_Batch(i,:))];
  r2_Adj_BR(i,1:4)=[nanmin(r2_Batch_adjusted(i,:)) nanmean(r2_Batch_adjusted(i,:)) nanmax(r2_Batch_adjusted(i,:)) nanstd(r2_Batch_adjusted(i,:))];
  SSE_BR(i,1:4)=[nanmin(r2_Batch_SSE(i,:)) nanmean(r2_Batch_SSE(i,:)) nanmax(r2_Batch_SSE(i,:)) nanstd(r2_Batch_SSE(i,:))];
  MSE_BR(i,1:4)=[nanmin(r2_Batch_MSE(i,:)) nanmean(r2_Batch_MSE(i,:)) nanmax(r2_Batch_MSE(i,:)) nanstd(r2_Batch_MSE(i,:))];
  RMSE_BR(i,1:4)=[nanmin(r2_Batch_RMSE(i,:)) nanmean(r2_Batch_RMSE(i,:)) nanmax(r2_Batch_RMSE(i,:)) nanstd(r2_Batch_RMSE(i,:))];     
 
  result_Batch=[R_BR;r2_Adj_BR;SSE_BR;MSE_BR;RMSE_BR];
  
  %% Online
  R_OR(i,1:4)=[nanmin(r_Online(i,:)) nanmean(r_Online(i,:)) nanmax(r_Online(i,:)) nanstd(r_Online(i,:))];
  r2_Adj_OR(i,1:4)=[nanmin(r2_Online_adjusted(i,:)) nanmean(r2_Online_adjusted(i,:)) nanmax(r2_Online_adjusted(i,:)) nanstd(r2_Online_adjusted(i,:))];
  SSE_OR(i,1:4)=[nanmin(r2_Online_SSE(i,:)) nanmean(r2_Online_SSE(i,:)) nanmax(r2_Online_SSE(i,:)) nanstd(r2_Online_SSE(i,:))];
  MSE_OR(i,1:4)=[nanmin(r2_Online_MSE(i,:)) nanmean(r2_Online_MSE(i,:)) nanmax(r2_Online_MSE(i,:)) nanstd(r2_Online_MSE(i,:))];
  RMSE_OR(i,1:4)=[nanmin(r2_Online_RMSE(i,:)) nanmean(r2_Online_RMSE(i,:)) nanmax(r2_Online_RMSE(i,:)) nanstd(r2_Online_RMSE(i,:))];     
 
  result_Online=[R_OR;r2_Adj_OR;SSE_OR;MSE_OR;RMSE_OR];
    
  %% Mini-Batch
  R_MBR(i,1:4)=[nanmin(r_MiniBatch(i,:)) nanmean(r_MiniBatch(i,:)) nanmax(r_MiniBatch(i,:)) nanstd(r_MiniBatch(i,:))];
  r2_Adj_MBR(i,1:4)=[nanmin(r2_MiniBatch_adjusted(i,:)) nanmean(r2_MiniBatch_adjusted(i,:)) nanmax(r2_MiniBatch_adjusted(i,:)) nanstd(r2_MiniBatch_adjusted(i,:))];
  SSE_MBR(i,1:4)=[nanmin(r2_MiniBatch_SSE(i,:)) nanmean(r2_MiniBatch_SSE(i,:)) nanmax(r2_MiniBatch_SSE(i,:)) nanstd(r2_MiniBatch_SSE(i,:))];
  MSE_MBR(i,1:4)=[nanmin(r2_MiniBatch_MSE(i,:)) nanmean(r2_MiniBatch_MSE(i,:)) nanmax(r2_MiniBatch_MSE(i,:)) nanstd(r2_MiniBatch_MSE(i,:))];
  RMSE_MBR(i,1:4)=[nanmin(r2_MiniBatch_RMSE(i,:)) nanmean(r2_MiniBatch_RMSE(i,:)) nanmax(r2_MiniBatch_RMSE(i,:)) nanstd(r2_MiniBatch_RMSE(i,:))];     
 
  result_MiniBatch=[R_MBR;r2_Adj_MBR;SSE_MBR;MSE_MBR;RMSE_MBR];
end

%% Finally, it is necessary to find the better results from learning method in the experimentation and plot it.
%% this should done for each learning method.
for i=1:out_targets
    %% Batch
    max_r_Batch=max(r_Batch(i,:));
    ind_r_Batch=find(r_Batch(i,:)== max_r_Batch);

    figure;
    plotregression(targets(i,:),outBatch_result{1,ind_r_Batch}{1,1}',strcat('Batch: X ',num2str(i,'%2d')))
    figure;
    plot(inputs(i,:),targets(i,:),'or',inputs(i,:),outBatch_result{1,ind_r_Batch}{1,1}','bo')
    title(strcat('Batch: X ',num2str(i,'%2d')))
            
    %% Plot the performance function progress
    figure('Name','Performance Progress');
    semilogy(1:size(trainBatch{1,ind_r_Batch}{1,1}.perf,2),trainBatch{1,ind_r_Batch}{1,1}.perf,'LineWidth',2); grid on
    xlabel('epochs'); ylabel('ln(SSE)');
    legend('Batch'); 
    
    %% Online
    max_r_Online=max(r_Online(i,:));
    ind_r_Online=find(r_Online(i,:)== max_r_Online);

    figure;
    plotregression(targets(i,:),outOnline_result{1,ind_r_Online}{1,1}',strcat('Online: X ',num2str(i,'%2d')))
    figure;
    plot(inputs(i,:),targets(i,:),'or',inputs(i,:),outOnline_result{1,ind_r_Online}{1,1}','bo')
    title(strcat('Online: X ',num2str(i,'%2d')))
      
        
    %% Plot the performance function progress
    figure('Name','Performance Progress');
    semilogy(1:size(trainOnline{1,ind_r_Online}{1,1}.perf,2),trainOnline{1,ind_r_Online}{1,1}.perf,'LineWidth',2); grid on
    xlabel('epochs'); ylabel('ln(SSE)');
    legend('Online');
    
    %% Mini-Batch
    max_r_MiniBatch=max(r_MiniBatch(i,:));
    ind_r_MiniBatch=find(r_MiniBatch(i,:)== max_r_MiniBatch);

    figure;
    plotregression(targets(i,:),outMiniBatch_result{1,ind_r_MiniBatch}{1,1}',strcat('Mini-Batch: X ',num2str(i,'%2d')))
    figure;
    plot(inputs(i,:),targets(i,:),'or',inputs(i,:),outMiniBatch_result{1,ind_r_MiniBatch}{1,1}','bo')
    title(strcat('MiniBatch: X ',num2str(i,'%2d')))
      
        
    %% Plot the performance function progress
    figure('Name','Performance Progress');
    semilogy(1:size(trainMiniBatch{1,ind_r_MiniBatch}{1,1}.perf,2),trainMiniBatch{1,ind_r_MiniBatch}{1,1}.perf,'LineWidth',2); grid on
    xlabel('epochs'); ylabel('ln(SSE)');
    legend('Mini-Batch');
end

