%% CC BY-SA (Attribution-ShareAlike) Mamdani based neuro-fuzzy system with center-of-sets defuzzification 
%by Dr. Juan R. Castro of Universidad Autnoma de Baja California 
%is distributed under a Creative Commons Attribution-ShareAlike 4.0 International License 
%(https://creativecommons.org/licenses/by-sa/4.0/). 
%This license allows others to remix, retouch, and create from their work, even for commercial purposes, 
%as long as they give credit and license their new creations under the same terms. All new works based on 
%yours will carry the same license, so any derivative works will also allow commercial use.

function fis = genfis1mamt1fnn(inputs, targets, rule_n , inmftype, outmftype)
% GENTSKT1FNN Generates an initial Mamdani-type FIS-COS for T1FNN training
%             using a noun grid partition.
%
%   FIS = GENMAMT1FNN(inputs,targets) generates a multiple-output Sugeno-type
%   fuzzy inference system (FIS) using a grid  partition on the data
%   (no clustering). FIS is used to provide initial conditions for TSKT1FNN
%   training. inputs is a matrix with N columns where N columns contain data
%   for each FIS input, and targets is a matrix with M columns where M contains
%   the output data. By default, GENTSKT1FNN uses two 'gaussmf' type membership
%   functions for each input. Each rule generated by GENTSKT1FNN has one output
%   membership function, which is of type 'linear' by default.
%
%   FIS = GENMAMT1FNN=(DATA, NUMMFS, INPUTMF, OUTPUTMF) explicitly specifies:
%     * NUMMFS   number of membership functions per input. A scalar value,
%                specifies the same number for all inputs and a vector value
%                specifies the number for each input individually.
%     * INPUTMF  type of membership function for each input. A single string
%                specifies the same type for all inputs, a string array
%                specifies the type for each input individually.
%     * OUTPUTMF output membership function type, either 'linear' or 'constant'
%
%   Example
%
%       load cho_dataset
%       inputs  = choInputs';
%       targets = choTargets';
%       fis = genmamt1fnn(inputs,targets,7,char('gaussmf'));
%

% Change this to have different default values
default_rule_n = 2;
default_mf_type = 'gaussmf';
default_output_type = 'constant';

if nargin <= 4,
	outmftype = default_output_type;
end
if nargin <= 3,
	inmftype = default_mf_type;
end
if nargin <= 2,
	rule_n = default_rule_n;
end

% get dimension info
[~,in_n] = size(inputs);
[~,out_n] = size(targets);
% error checking
numMFs=rule_n*ones(1, in_n);

% Check arguments defining system inputs
if length(numMFs) ~= in_n | (size(inmftype, 1) ~=1 & size(inmftype, 1) ~= in_n),
	error('Wrong size(s) of argument(s) defining system input(s)!');
end
% Check argument defining system output
if size(outmftype,1) ~= 1
	error('Argument data entered may only have one output!');
end
if (strcmp(outmftype,'linear') | strcmp(outmftype,'constant')) ~= 1
    error('Output membership function type must be either linear or constant!');
end

if size(inmftype, 1) ==1 &  in_n>1
   for i=2:in_n
      inmftype(i,:)=inmftype(1,:);
   end
end

fis.name = 'mamt1fnn';
fis.type = 'sugeno';
fis.andMethod = 'prod';
fis.orMethod  = 'max';
fis.defuzzMethod = 'wtaver';
fis.impMethod = 'prod';
fis.aggMethod = 'max';

in_range = [min(inputs,[],1)' max(inputs,[],1)'];
in_mf_param = genparamt1fnn(inputs, numMFs, inmftype);
k=1;
for i = 1:in_n,
    fis.input(i).name = ['input' num2str(i)];
    fis.input(i).range=in_range(i,:);
    for j=1:numMFs(i)
        MFType = deblank(inmftype(i, :));
        fis.input(i).mf(j).name = ['in' num2str(i) 'mf' num2str(j)];
        fis.input(i).mf(j).type = MFType;
        if strcmp(MFType,'gaussmf') | strcmp(MFType,'sigmf') ...
                    | strcmp(MFType,'smf'),
            fis.input(i).mf(j).params= in_mf_param(k,1:2);
        elseif strcmp(MFType,'trimf') | strcmp(MFType,'gbellmf'),
            fis.input(i).mf(j).params= in_mf_param(k,1:3);
        else
            fis.input(i).mf(j).params= in_mf_param(k,1:4);
        end  
        k=k+1;
    end
end

out_range = [min(targets,[],1)' max(targets,[],1)'];
for j = 1:out_n,
    fis.output(j).name = ['output' num2str(j)];
    fis.output(j).range= out_range(j,:);
    for k = 1:rule_n,        
        fis.output(j).mf(k).name=['out' num2str(k) 'mf' num2str(j)];  
        fis.output(j).mf(k).type=outmftype;
        if strcmp(outmftype, 'linear')
            fis.output(j).mf(k).params=zeros(1,in_n+1);
        else
        fis.output(j).mf(k).params=[0];
        end
    end
end

rule_list = repmat((1:rule_n)',[1 in_n+out_n]);
rule_list = [rule_list ones(rule_n, 1) ones(rule_n, 1)];
fis.rule=[];
fis=setfis(fis, 'rulelist', rule_list);
