%% CC BY-SA (Attribution-ShareAlike) Mamdani based neuro-fuzzy system with center-of-sets defuzzification 
%by Dr. Juan R. Castro of Universidad Autnoma de Baja California 
%is distributed under a Creative Commons Attribution-ShareAlike 4.0 International License 
%(https://creativecommons.org/licenses/by-sa/4.0/). 
%This license allows others to remix, retouch, and create from their work, even for commercial purposes, 
%as long as they give credit and license their new creations under the same terms. All new works based on 
%yours will carry the same license, so any derivative works will also allow commercial use.

function mf_param = genparamt1fnn(data,mf_n,mf_type)
%   GENPARAMTSKT1FNN Generate initial membership function parameters for MAMT1FNN learning.
default_mf_n = 2;
default_mf_type = 'gaussmf';

if nargin <= 2,
    mf_type = default_mf_type;
end
if nargin <= 1,
    mf_n = default_mf_n;
end

% get dimension info
% data_n = size(data, 1);
in_n = size(data, 2);
range = [min(data,[],1)' max(data,[],1)'];

% generate mf_n and mf_type of proper sizes
if length(mf_n) == 1,
    mf_n = mf_n(ones(in_n, 1), :);
end
if size(mf_type, 1) == 1,   % single mf_type for all MFs
    mf_type = mf_type(ones(in_n, 1), :);
end

% error checking
if length(mf_n) ~= in_n | size(mf_type, 1) ~= in_n,
    error('Wrong sizes of given argument(s)!');
end

mf_param = [];
for i = 1:in_n;
    type = deblank(mf_type(i, :));
    if strcmp(type, 'gbellmf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        b = 2;
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        d = zeros(mf_n(i), 1);
        mf_param = [mf_param; 
            a(ones(mf_n(i), 1)) b(ones(mf_n(i), 1)) c d];
    elseif strcmp(type, 'gaussmf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        a = a/sqrt(2*log(2));
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        d = zeros(mf_n(i), 1);
        mf_param = [mf_param; a(ones(mf_n(i), 1)) c d d];
    elseif strcmp(type, 'gauss2mf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        c1 = c - 0.6*a;
        c2 = c + 0.6*a;
        sigma = 0.4*a/sqrt(2*log(2))*ones(mf_n(i), 1); 
        mf_param = [mf_param; sigma c1 sigma c2];
    elseif strcmp(type, 'dsigmf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        tmp = 10/(range(i,2) - range(i, 1))*mf_n(i)*ones(mf_n(i), 1);
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        mf_param = [mf_param; tmp c-a tmp c+a];
    elseif strcmp(type, 'psigmf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        tmp = 10/(range(i,2) - range(i, 1))*mf_n(i)*ones(mf_n(i), 1);
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        mf_param = [mf_param; tmp c-a -tmp c+a];
    elseif strcmp(type, 'trimf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        d = zeros(mf_n(i), 1);
        mf_param = [mf_param; c-2*a c c+2*a d];
    elseif strcmp(type, 'trapmf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        d = zeros(mf_n(i), 1);
        mf_param = [mf_param; c-1.4*a c-0.6*a c+0.6*a c+1.4*a];
    elseif strcmp(type, 'pimf'),
        a = (range(i,2) - range(i, 1))/2/(mf_n(i) - 1);
        c = linspace(range(i, 1), range(i, 2), mf_n(i))';
        mf_param = [mf_param; c-1.4*a c-0.6*a c+0.6*a c+1.4*a];
    else
        fprintf('mf_type = %s\n', type);
        error('Unsupported MF type!');
    end
end
