%% CC BY-SA (Attribution-ShareAlike) Mamdani based neuro-fuzzy system with center-of-sets defuzzification 
%by Dr. Juan R. Castro of Universidad Autnoma de Baja California 
%is distributed under a Creative Commons Attribution-ShareAlike 4.0 International License 
%(https://creativecommons.org/licenses/by-sa/4.0/). 
%This license allows others to remix, retouch, and create from their work, even for commercial purposes, 
%as long as they give credit and license their new creations under the same terms. All new works based on 
%yours will carry the same license, so any derivative works will also allow commercial use.

%This code was based on traingdx deployed by Mark Beale under the follow license:
% ODJ, 11/20/98, added support for user stopping. 
% Copyright 1992-2002 The MathWorks, Inc. 
% $Revision: 1.10 $ $Date: 2002/04/14 21:35:41 $
%The contribution added by us in this code are the functions that calculate the error signals and gradient vector, 
%these are: mamcost1fnncalcperf and mamcost1fnncalcgrad respectively.


function [fis,tr] = mamcost1fnntraingdx_Batch(fis,trainV,valV,testV,trainParam)
% MAMCOST1FNNTRAINGDXBATCH Gradient descent w/momentum & adaptive lr backpropagation.
%
%	Syntax
%	
%   [fis,tr] = MAMCOST1FNNTRAINGDXBATCH(fis,trainV,valV,testV,trainParam)
%
%	Description
%
%	  MAMCOST1FNNTRAINGDX is a mamcostfnn training function that updates weight values
%     according to gradient descent momentum and an adaptive learning rate.
%
%	  Training occurs according to the TSKT1FNNTRAINGDX's training parameters
%	  shown here with their default values:
%
%	    trainParam.epochs         10  Maximum number of epochs to train
%	    trainParam.goal            0  Performance goal
%	    trainParam.lr           0.01  Learning rate
%	    trainParam.lr_inc       1.05  Ratio to increase learning rate
%	    trainParam.lr_dec        0.7  Ratio to decrease learning rate
%	    trainParam.max_fail        5  Maximum validation failures
%	    trainParam.max_perf_inc 1.04  Maximum performance increase
%	    trainParam.mc            0.9  Momentum constant.
%	    trainParam.min_grad    1e-10  Minimum performance gradient
%	    trainParam.show           25  Epochs between showing progress
%	    trainParam.time          inf  Maximum time to train in seconds
%
%	Algorithm
%
%	  MAMCOST1FNNTRAINGDX can train any tskt1fnn as long as its weight, net input,
%	  and membership functions have derivative functions.
%
%	  Backpropagation is used to calculate derivatives of performance
%	  PERF with respect to the weight variables X.  Each
%	  variable is adjusted according to the gradient descent
%	  with momentum.
%
%	    dX = mc*dXprev + lr*mc*dperf/dX
%
%	  where dXprev is the previous change to the weight.
%
%	  For each epoch, if performance decreases toward the goal, then
%	  the learning rate is increased by the factor lr_inc.  If
%	  performance increases by more than the factor max_perf_inc,
%	  the learning rate is adjusted by the factor lr_dec and the
%	  change, which increased the performance, is not made.
%
%	  Training stops when any of these conditions occur:
%	  1) The maximum number of EPOCHS (repetitions) is reached.
%	  2) The maximum amount of TIME has been exceeded.
%	  3) Performance has been minimized to the GOAL.
%	  4) The performance gradient falls below MINGRAD.
%	  5) Validation performance has increase more than MAX_FAIL times
%	     since the last time it decreased (when using validation).
%
%
% FUNCTION INFO
% =============
%     pdefaults:
%
% 	  trainParam.epochs = 100;
% 	  trainParam.goal = 0;
% 	  trainParam.lr = 0.01;
% 	  trainParam.lr_dec = 0.7;
% 	  trainParam.lr_inc = 1.05;
% 	  trainParam.max_fail = 5;
% 	  trainParam.max_perf_inc = 1.04;
% 	  trainParam.mc = 0.9;
% 	  trainParam.min_grad = 1.0e-6;
% 	  trainParam.show = 25;
% 	  trainParam.time = inf;

% CALCULATION
% ===========
%
% Constants
this = 'MAMCOST1FNNTRAINGDX_BATCH';
performFcn = 'sse';

epochs = trainParam.epochs;
goal = trainParam.goal;
lr = trainParam.lr;
lr_inc = trainParam.lr_inc;
lr_dec = trainParam.lr_dec;
max_fail = trainParam.max_fail;
max_perf_inc = trainParam.max_perf_inc;
mc = trainParam.mc;
min_grad = trainParam.min_grad;
show = trainParam.show;
time = trainParam.time;
experiment=trainParam.experiment;

doValidation = ~isempty(valV.indices);
doTest = ~isempty(testV.indices);

% Initialize
stop = '';
startTime = clock;
X = fnngetx(fis);
[perf,~,E,A] = mamcost1fnncalcperf(trainV.X,trainV.T,fis);
[gX,normgX] =mamcost1fnncalcgrad(fis,trainV.X,E,A);
dX = -lr*gX;

if (doValidation)
  VV.fis = fis;
  VV.perf = mamcost1fnncalcperf(valV.X,valV.T,fis);
  VV.numFail = 0;
end

tr = newtr(epochs,'perf','vperf','tperf','lr');

last_perf_trained=0.0;
epochLast=0.0;
countEqual=0.0;

% Train
for epoch=0:epochs
    
% Gradient Descent with Momentum and Adaptive Learning Rate
  dX = mc*dX - (1-mc)*lr*gX;
  X2 = X + dX;
  fis2 = fnnsetx(fis,X2);
  [perf2,~,E2,A2] = mamcost1fnncalcperf(trainV.X,trainV.T,fis2);
  
  if (perf2/perf) > max_perf_inc
    lr = lr*lr_dec;
	dX = lr*gX;
  else
      
     if (perf2 < perf)
       lr = lr*lr_inc;
     end
     
    X = X2;
	fis = fis2;
	perf = perf2;

    [gX,normgX] = mamcost1fnncalcgrad(fis,trainV.X,E2,A2);
  end
  
  if last_perf_trained~=0 && perf==last_perf_trained && (epoch-epochLast)==1
      countEqual=countEqual+1;
      
      if countEqual==1000
          fprintf(' CountEqual: %g , Perf: %g, last_perf_trained: %g, NO CHANGE',countEqual,perf,last_perf_trained);
          break;
      end
  else 
      countEqual=0;
  end    
  
  if isnan(perf)
    fprintf(' Perf: %g IS EQUAL TO NAN',perf);  
    break;
  end
  
  last_perf_trained=perf;
  epochLast=epoch;
  
  % Training Record
  epochPlus1 = epoch+1;
  tr.perf(epochPlus1) = perf;
  tr.lr(epochPlus1) = lr;
 
  
  if (doValidation)
    tr.vperf(epochPlus1) = VV.perf;
  end
  
  if (doTest)
    tr.tperf(epochPlus1) = mamcost1fnncalcperf(testV.X,testV.T,fis);
  end
  
  % Stopping Criteria
  currentTime = etime(clock,startTime);
  if (perf <= goal)
    stop = 'Performance goal met.';
  elseif (epoch == epochs)
    stop = 'Maximum epoch reached, performance goal was not met.';
  elseif (currentTime > time)
    stop = 'Maximum time elapsed, performance goal was not met.';
  elseif (normgX < min_grad)
    stop = 'Minimum gradient reached, performance goal was not met.';
  elseif (doValidation) & (VV.numFail > max_fail)
    stop = 'Validation stop.';
	fis = VV.fis;
  end
  % Progress
  if ~rem(epoch,show) | length(stop)
    fprintf(this);
	if isfinite(epochs) fprintf(', Experiment: %g  Epoch %g/%g',experiment,epoch, epochs); end
	if isfinite(time) fprintf(', Time %g%%',currentTime/time/100); end
	if isfinite(goal) fprintf(', %s %g/%g',upper(performFcn),perf,goal); end
	if isfinite(min_grad) fprintf(', Gradient %g/%g',normgX,min_grad); end
	fprintf('\n');
	plotperf(tr,goal,this,epoch);
    if length(stop) fprintf('%s, %s\n\n',this,stop); break; end
  end
  
  % Validation
  if (doValidation)
    vperf = mamcost1fnncalcperf(valV.X,valV.T,fis);
	if (vperf < VV.perf)
	  VV.perf = vperf; VV.fis = fis; VV.numFail = 0;
	elseif (vperf > VV.perf)
      VV.numFail = VV.numFail + 1;
	end
  end
  
end

% Finish
tr = cliptr(tr,epoch);
